SystemOrganization addCategory: #LDAS!

Object subclass: #LDAllocator
	instanceVariableNames: 'topicCount featureCount documentCount topicPriors perDocumentTopics perDocumentFrequency perTopicFeatures perTopicFrequency perWordTopic featurePriors documents'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LDAS'!

!LDAllocator commentStamp: 'toma 8/6/2015 01:08' prior: 0!
LDAllocator trains a Latent Dirichlet Allocation (LDA) Model on a given set of documents. It takes a set of documents (vectors of features) as input, clusters co-occuring features into (latent) topics and assigns each document an affinity towards each topic.

Documents need to be represented as sequences of natural numbers. Example:

	sixDocs := #(  #(1 2 3)  #(1 2 2 2)  #(1 1 3)  #(3 4 5) #(4 5) #(3 5 4 3) )

	(You can use LDFeatureSpace to map any hashable type to such lists)

We know we have 5 distinct features (1 to 5) and want to have them arranged in two topics:

	lda := LDAllocator new.
	model := lda allocate: sixDocs iterating: 10 features: 5topics: 2.

Let's inspect the topics:

	model featuresForTopicAt: 1
		#(4.0 4.0 2.0 1.0 0.0)
		which means features 1 and 2 are in that topic, 3 and 4 slightly.
	model featuresForTopicAt: 2
		 #(0.0 0.0 3.0 2.0 3.0)
		which means features 3, 4, and 5 reside in this topic.

Let's inspect which topics a document belongs to:

	model topicsForDocumentAt: 6
		#(0.0 4.0)
		which shows strong correspondence to topic 2 (containing 3, 4, and 5 as features).

Let's compute the similarity between two documents based on what we know about their topics:

	model documentSimilarityOf: 1 and: 2
		12.0
	model documentSimilarityOf: 5 and: 6
		4.0
	model documentSimilarityOf: 1 and: 6
		0.0

	which indicates that documents 1 and 2 are concerned with the same thing, 5 and 6 approximately with the same thing, and 1 and 6 are different concerns.
	



!

!LDAllocator methodsFor: 'as yet unclassified' stamp: 'toma 8/6/2015 00:29'!
allocate: docs iterating: numIterations features: numFeatures topics: numTopics
	"Train model with given documents for given number of iterations."

	self setupForDocuments: docs
			featureCount: numFeatures
			topicCount: numTopics;
		randomize.
	(1 to: numIterations) do: [ :i | self update].
	^self model
	
	! !

!LDAllocator methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 16:51'!
associateFeature: feature withDoc: doc andTopic: topic
	
	self associateFeature: feature withDoc: doc andTopic: topic by: 1.! !

!LDAllocator methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 18:36'!
associateFeature: feature withDoc: doc andTopic: topic by: delta
	"update statistics to reflect association between word, topic and document"

	perDocumentTopics at: doc at: topic put: 
		(delta + (perDocumentTopics at: doc at: topic)).

	perTopicFeatures at: topic at: feature put: 
		(delta + (perTopicFeatures at: topic at: feature)).

	perDocumentFrequency at: doc put:
		(delta + (perDocumentFrequency at: doc)).

	perTopicFrequency at: topic put:
		(delta + (perTopicFrequency at: topic)).! !

!LDAllocator methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 16:51'!
dissociateFeature: feature withDoc: doc andTopic: topic
	
	self associateFeature: feature withDoc: doc andTopic: topic by: (1 negated).! !

!LDAllocator methodsFor: 'as yet unclassified' stamp: 'toma 8/14/2015 18:37'!
featurePriors
	^featurePriors! !

!LDAllocator methodsFor: 'as yet unclassified' stamp: 'toma 8/14/2015 18:37'!
featurePriors: aCollection
	featurePriors := aCollection! !

!LDAllocator methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 16:20'!
initialize
	! !

!LDAllocator methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 18:55'!
model
	^(LDModel new)
			initializeTopics: perTopicFeatures 
			andDocuments: perDocumentTopics 
			priors: topicPriors;
		yourself.! !

!LDAllocator methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 16:51'!
randomize
	"Initialize the topic model by assigning a random topic to each word"

	| topic  random |

	random := Random new.
	
	documents withIndexDo: [ :doc :di |
		doc withIndexDo: [ :word :wj |
			topic := random nextInt: topicCount.
			self associateFeature: word withDoc: di andTopic: topic. 	
			perWordTopic at: (di @ wj) put: topic.	
		]
	]! !

!LDAllocator methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 16:32'!
setupForDocuments: docs featureCount: features topicCount: topics
	
	documents := docs.
	topicCount := topics.
	featureCount := features.
	documentCount := docs size.
	
	perDocumentTopics := Matrix
		rows: documentCount 
		columns: topicCount
		element: 0.0.
	perDocumentFrequency := (Array new: documentCount) atAllPut: 0.
	perTopicFeatures := Matrix
		rows: topicCount
		columns: featureCount
		element: 0.0.
	perTopicFrequency := (Array new: topicCount) atAllPut: 0.

	topicPriors := (Array new: topicCount) atAllPut: (1.0 / topicCount).
	featurePriors := (Array new: featureCount) atAllPut: (1.0 / featureCount).

	perWordTopic := Dictionary new.

	


	
	! !

!LDAllocator methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 18:08'!
topicProbabilityOver: doc andFeature: feature
	"Compute P(t | d, f) given previous topic allocation and priors.
	The result is a distribution, not a value!!"

	| featurePrior topics docFrequency |

	featurePrior := featurePriors at: feature.
	topics := perDocumentTopics atRow: doc.
	docFrequency := perDocumentFrequency at: doc.

	^ LDMultinomial normalized: (
		(1 to: topicCount) collect: [ :topic | | pFeature pTopic |

		"bayesian estimates proportional to feature and topic marginals
		(with normalization terms, don't ask!!)"

		pFeature := (perTopicFeatures at: topic at: feature) + featurePrior.		
		pFeature := pFeature / (1.0 + (perTopicFrequency at: topic)).

		pTopic := (topics at: topic) + (topicPriors at: topic).
		pTopic := pTopic / (docFrequency + ((topicPriors at: topic) * topicCount)).

		pFeature * pTopic		
	])! !

!LDAllocator methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 18:16'!
update
	| topic | 

	documents withIndexDo: [:document :di |
		document withIndexDo: [:word :wj |
			topic := perWordTopic at: (di @ wj).
			self dissociateFeature: word withDoc: di andTopic: topic.
			topic := (self topicProbabilityOver: di 
 							andFeature: word) sample.
			self associateFeature: word withDoc: di andTopic: topic.
			perWordTopic at: (di @ wj) put: topic.
		]
	]! !

LDAllocator subclass: #LDContextAllocator
	instanceVariableNames: 'contextWeight contextField'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LDAS'!

!LDContextAllocator methodsFor: 'as yet unclassified' stamp: 'toma 8/14/2015 14:31'!
blanketOf: aDocId at: aPosition
	
	^contextField markovBlanket: (aDocId @ aPosition)! !

!LDContextAllocator methodsFor: 'accessing' stamp: 'toma 8/14/2015 14:47'!
contextField
	^ contextField! !

!LDContextAllocator methodsFor: 'accessing' stamp: 'toma 8/14/2015 14:47'!
contextField: anObject
	contextField := anObject! !

!LDContextAllocator methodsFor: 'as yet unclassified' stamp: 'toma 8/14/2015 14:37'!
contextTopicsOf: aDocId at: aPosition
	
	^ (self blanketOf: aDocId at: aPosition)
		collect: [ :neighbor | perWordTopic at: neighbor]! !

!LDContextAllocator methodsFor: 'accessing' stamp: 'toma 8/14/2015 16:19'!
contextWeight
	^ contextWeight! !

!LDContextAllocator methodsFor: 'accessing' stamp: 'toma 8/14/2015 16:19'!
contextWeight: anObject
	contextWeight := anObject! !

!LDContextAllocator methodsFor: 'accessing' stamp: 'toma 8/14/2015 16:19'!
documentCount
	^ documentCount! !

!LDContextAllocator methodsFor: 'accessing' stamp: 'toma 8/14/2015 16:19'!
documentCount: anObject
	documentCount := anObject! !

!LDContextAllocator methodsFor: 'accessing' stamp: 'toma 8/14/2015 16:19'!
documents
	^ documents! !

!LDContextAllocator methodsFor: 'accessing' stamp: 'toma 8/14/2015 16:19'!
documents: anObject
	documents := anObject! !

!LDContextAllocator methodsFor: 'accessing' stamp: 'toma 8/14/2015 16:19'!
featureCount
	^ featureCount! !

!LDContextAllocator methodsFor: 'accessing' stamp: 'toma 8/14/2015 16:19'!
featureCount: anObject
	featureCount := anObject! !

!LDContextAllocator methodsFor: 'accessing' stamp: 'toma 8/14/2015 16:19'!
featurePriors
	^ featurePriors! !

!LDContextAllocator methodsFor: 'accessing' stamp: 'toma 8/14/2015 16:19'!
featurePriors: anObject
	featurePriors := anObject! !

!LDContextAllocator methodsFor: 'as yet unclassified' stamp: 'toma 8/14/2015 14:46'!
initialize 
	
	super initialize.
	contextWeight := 0.5.! !

!LDContextAllocator methodsFor: 'accessing' stamp: 'toma 8/14/2015 16:19'!
perDocumentFrequency
	^ perDocumentFrequency! !

!LDContextAllocator methodsFor: 'accessing' stamp: 'toma 8/14/2015 16:19'!
perDocumentFrequency: anObject
	perDocumentFrequency := anObject! !

!LDContextAllocator methodsFor: 'accessing' stamp: 'toma 8/14/2015 16:19'!
perDocumentTopics
	^ perDocumentTopics! !

!LDContextAllocator methodsFor: 'accessing' stamp: 'toma 8/14/2015 16:19'!
perDocumentTopics: anObject
	perDocumentTopics := anObject! !

!LDContextAllocator methodsFor: 'accessing' stamp: 'toma 8/14/2015 16:19'!
perTopicFeatures
	^ perTopicFeatures! !

!LDContextAllocator methodsFor: 'accessing' stamp: 'toma 8/14/2015 16:19'!
perTopicFeatures: anObject
	perTopicFeatures := anObject! !

!LDContextAllocator methodsFor: 'accessing' stamp: 'toma 8/14/2015 16:19'!
perTopicFrequency
	^ perTopicFrequency! !

!LDContextAllocator methodsFor: 'accessing' stamp: 'toma 8/14/2015 16:19'!
perTopicFrequency: anObject
	perTopicFrequency := anObject! !

!LDContextAllocator methodsFor: 'accessing' stamp: 'toma 8/14/2015 16:19'!
perWordTopic
	^ perWordTopic! !

!LDContextAllocator methodsFor: 'accessing' stamp: 'toma 8/14/2015 16:19'!
perWordTopic: anObject
	perWordTopic := anObject! !

!LDContextAllocator methodsFor: 'accessing' stamp: 'toma 8/14/2015 16:19'!
topicCount
	^ topicCount! !

!LDContextAllocator methodsFor: 'accessing' stamp: 'toma 8/14/2015 16:19'!
topicCount: anObject
	topicCount := anObject! !

!LDContextAllocator methodsFor: 'accessing' stamp: 'toma 8/14/2015 16:19'!
topicPriors
	^ topicPriors! !

!LDContextAllocator methodsFor: 'accessing' stamp: 'toma 8/14/2015 16:19'!
topicPriors: anObject
	topicPriors := anObject! !

!LDContextAllocator methodsFor: 'as yet unclassified' stamp: 'toma 8/14/2015 16:21'!
update
	| topic | 

	documents withIndexDo: [:document :di |
		document withIndexDo: [:word :wj |
			topic := perWordTopic at: (di @ wj).
			self dissociateFeature: word withDoc: di andTopic: topic.
			((LDMultinomial rand next > contextWeight) and: [document size > 1])
				ifTrue:
					[topic := (self contextTopicsOf: di at: wj) atRandom ]
				ifFalse: 
					[topic := (self topicProbabilityOver: di 
 						andFeature: word) sample].
			self associateFeature: word withDoc: di andTopic: topic.
			perWordTopic at: (di @ wj) put: topic.
		]
	]! !

LDAllocator subclass: #LDLinkedAllocator
	instanceVariableNames: 'contextWeight linkField'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LDAS'!

!LDLinkedAllocator methodsFor: 'accessing' stamp: 'toma 8/14/2015 16:54'!
contextWeight
	^ contextWeight! !

!LDLinkedAllocator methodsFor: 'accessing' stamp: 'toma 8/14/2015 16:54'!
contextWeight: anObject
	contextWeight := anObject! !

!LDLinkedAllocator methodsFor: 'as yet unclassified' stamp: 'toma 8/14/2015 17:02'!
initialize 
	
	super initialize.
	contextWeight := 0.5.
	linkField := LDLinkField new.! !

!LDLinkedAllocator methodsFor: 'accessing' stamp: 'toma 8/14/2015 16:54'!
linkField
	^ linkField! !

!LDLinkedAllocator methodsFor: 'accessing' stamp: 'toma 8/14/2015 16:54'!
linkField: anObject
	linkField := anObject! !

!LDLinkedAllocator methodsFor: 'as yet unclassified' stamp: 'toma 8/14/2015 17:28'!
peerMixtureOf: document at: position

	^((self linkField linksFromDocument: document at: position)
		inject: (LDNullMultinomial new)
		into: [:state :doc | state ++ (LDMultinomial raw: (perDocumentTopics atRow: doc))]) normalize.! !

!LDLinkedAllocator methodsFor: 'as yet unclassified' stamp: 'toma 8/14/2015 18:30'!
sampleFromPeersOf: document at: position

	| peer peerPosition |

	peer := (self linkField linksFromDocument: document at: position) atRandom.
	peerPosition := LDMultinomial rand nextInt: (documents at: peer) size.
	^perWordTopic at: (peer@peerPosition)! !

!LDLinkedAllocator methodsFor: 'as yet unclassified' stamp: 'toma 8/14/2015 18:30'!
update
	| topic | 

	documents withIndexDo: [:document :di |
		document withIndexDo: [:word :wj |
			topic := perWordTopic at: (di @ wj).
			self dissociateFeature: word withDoc: di andTopic: topic.

			((LDMultinomial rand next > contextWeight) and: [self linkField isDocument: di linkedAt: wj])
				ifTrue:
					"[topic := (self peerMixtureOf: di at: wj) sample"
					[topic := self sampleFromPeersOf: di at: wj]
				ifFalse: 
					[topic := (self topicProbabilityOver: di 
 						andFeature: word) sample].

			self associateFeature: word withDoc: di andTopic: topic.
			perWordTopic at: (di @ wj) put: topic.
		]
	]! !

Object subclass: #LDConceptAllocator
	instanceVariableNames: 'documentPairs topicCount featureCount mixturePrior featurePrior relationPrior docMixtures topicMixtures topicMarginals translations translationMarginals directTopics externalTopics internalTopics'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LDAS'!

Object subclass: #LDDocumentPair
	instanceVariableNames: 'conceptFeatures implementationFeatures'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LDAS'!

Object subclass: #LDFeatureSpace
	instanceVariableNames: 'featureIdMap idFeatureMap maxId'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LDAS'!

!LDFeatureSpace commentStamp: 'toma 8/6/2015 00:51' prior: 0!
This class helps to transform arbitrary hashable features (e.g. words or collections of words) into natural numbers. The transformation is required, as LDAllocator only accepts integer features.

Example:

	space := LDFeatureSpace new.
	space representAll: #(a b c 100).
		gives #(1 2 3 4)
	space representAll: #(a c d).
		subsequently gives  #(1 3 5) because a and c are already known as feature #1 and #3, c is inserted as new feature #5.

The space needs to be retained for later interpretation of encoded results:

	space interpret: 2
		gives #b
	space interpretAll: #(5 3 2 1)
		gives #(#d #c #b #a)
!

LDFeatureSpace subclass: #LDDocumentFeatureSpace
	instanceVariableNames: 'documentCount perFeatureDocuments perFeatureFrequency perDocumentSize documentFeatureCounts'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LDAS'!

!LDDocumentFeatureSpace methodsFor: 'as yet unclassified' stamp: 'toma 8/14/2015 16:01'!
TfIdf: aFeatureId in: aDocumentId

	^ (self frequencyOf: aFeatureId in: aDocumentId) 
		* ((self documentCount / (self perFeatureDocuments at: aFeatureId)) log)! !

!LDDocumentFeatureSpace methodsFor: 'as yet unclassified' stamp: 'toma 8/14/2015 15:51'!
addNewFeature
	self perFeatureDocuments add: 0.
	self perFeatureFrequency add: 0.! !

!LDDocumentFeatureSpace methodsFor: 'as yet unclassified' stamp: 'toma 8/14/2015 15:41'!
associateFeature: aFeatureId with: aDocumentId

	| count |
	count := self documentFeatureCounts
		at: (aDocumentId @ aFeatureId)
		ifAbsentPut: [0].
	self documentFeatureCounts
		at: (aDocumentId @ aFeatureId)
		put: (1 + count).! !

!LDDocumentFeatureSpace methodsFor: 'accessing' stamp: 'toma 8/14/2015 15:01'!
documentCount
	^ documentCount! !

!LDDocumentFeatureSpace methodsFor: 'accessing' stamp: 'toma 8/14/2015 15:01'!
documentCount: anObject
	documentCount := anObject! !

!LDDocumentFeatureSpace methodsFor: 'accessing' stamp: 'toma 8/14/2015 15:03'!
documentFeatureCounts
	^ documentFeatureCounts! !

!LDDocumentFeatureSpace methodsFor: 'accessing' stamp: 'toma 8/14/2015 15:03'!
documentFeatureCounts: anObject
	documentFeatureCounts := anObject! !

!LDDocumentFeatureSpace methodsFor: 'as yet unclassified' stamp: 'toma 8/14/2015 16:04'!
documentProportionOf: aFeatureId

	^ (self perFeatureFrequency at: aFeatureId) / (self documentCount)! !

!LDDocumentFeatureSpace methodsFor: 'as yet unclassified' stamp: 'toma 8/14/2015 18:41'!
estimateFeaturePriors

	| sum |
	sum := perFeatureFrequency sum.	
	^perFeatureFrequency collect: [ :freq | (freq / sum) asFloat ] ! !

!LDDocumentFeatureSpace methodsFor: 'as yet unclassified' stamp: 'toma 8/14/2015 15:55'!
frequencyOf: aFeature in: aDocument

	^ self documentFeatureCounts at: (aDocument @ aFeature) ifAbsent: 0.! !

!LDDocumentFeatureSpace methodsFor: 'as yet unclassified' stamp: 'toma 8/14/2015 15:06'!
initialize 
	super initialize.
	documentCount := 0.
	documentFeatureCounts := Dictionary new.
	perDocumentSize := OrderedCollection new.
	perFeatureDocuments := OrderedCollection new.
	perFeatureFrequency := OrderedCollection new.! !

!LDDocumentFeatureSpace methodsFor: 'accessing' stamp: 'toma 8/14/2015 15:01'!
perDocumentSize
	^ perDocumentSize! !

!LDDocumentFeatureSpace methodsFor: 'accessing' stamp: 'toma 8/14/2015 15:01'!
perDocumentSize: anObject
	perDocumentSize := anObject! !

!LDDocumentFeatureSpace methodsFor: 'accessing' stamp: 'toma 8/14/2015 15:01'!
perFeatureDocuments
	^ perFeatureDocuments! !

!LDDocumentFeatureSpace methodsFor: 'accessing' stamp: 'toma 8/14/2015 15:01'!
perFeatureDocuments: anObject
	perFeatureDocuments := anObject! !

!LDDocumentFeatureSpace methodsFor: 'accessing' stamp: 'toma 8/14/2015 15:01'!
perFeatureFrequency
	^ perFeatureFrequency! !

!LDDocumentFeatureSpace methodsFor: 'accessing' stamp: 'toma 8/14/2015 15:01'!
perFeatureFrequency: anObject
	perFeatureFrequency := anObject! !

!LDDocumentFeatureSpace methodsFor: 'as yet unclassified' stamp: 'toma 8/14/2015 15:52'!
represent: aFeature

	| id |
	id := super represent: aFeature.
	self perFeatureFrequency at: id put: (1 + self perFeatureFrequency at: id).
	^id! !

!LDDocumentFeatureSpace methodsFor: 'as yet unclassified' stamp: 'toma 8/14/2015 15:46'!
representAll: aCollection

	| doc uniqueFeatures featureVector |
	
	doc := self documentCount + 1.
	self documentCount: doc.

	uniqueFeatures := Set new.

	featureVector := aCollection collect: [ :each | | id | 
		id := self represent: each.
		self associateFeature: id with: doc. 
		uniqueFeatures add: id.
		id].

	uniqueFeatures do: [ :id |
		self perFeatureDocuments at: id put: (1 + (self perFeatureDocuments at: id))].
	
	self perDocumentSize add: (aCollection size).
	^featureVector! !

!LDFeatureSpace methodsFor: 'as yet unclassified' stamp: 'toma 8/14/2015 15:51'!
addNewFeature

	"Hook for subclasses to reserve storage for new feature"! !

!LDFeatureSpace methodsFor: 'as yet unclassified' stamp: 'toma 8/6/2015 00:34'!
featureCount
	
	^maxId! !

!LDFeatureSpace methodsFor: 'as yet unclassified' stamp: 'toma 8/6/2015 00:33'!
initialize 
	featureIdMap := Dictionary new.
	idFeatureMap := Dictionary new.
	maxId := 0.! !

!LDFeatureSpace methodsFor: 'as yet unclassified' stamp: 'toma 8/6/2015 00:39'!
interpret: aFeatureId
	
	^idFeatureMap at: aFeatureId! !

!LDFeatureSpace methodsFor: 'as yet unclassified' stamp: 'toma 8/6/2015 00:41'!
interpretAll: aCollection
	
	^aCollection collect: [ :each | self interpret: each ]! !

!LDFeatureSpace methodsFor: 'as yet unclassified' stamp: 'toma 8/14/2015 15:51'!
represent: aFeature
	
	^featureIdMap at: aFeature ifAbsentPut: [ 
		maxId := maxId + 1.
		idFeatureMap at: maxId put: aFeature.
		self addNewFeature.
		maxId]! !

!LDFeatureSpace methodsFor: 'as yet unclassified' stamp: 'toma 8/6/2015 00:40'!
representAll: aCollection

	^aCollection collect: [ :each | self represent: each ]! !

Object subclass: #LDLinkField
	instanceVariableNames: 'adjacencyLists'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LDAS'!

!LDLinkField methodsFor: 'as yet unclassified' stamp: 'toma 8/14/2015 16:58'!
initialize

	adjacencyLists := Dictionary new.! !

!LDLinkField methodsFor: 'as yet unclassified' stamp: 'toma 8/14/2015 17:28'!
isDocument: document linkedAt: position

	^adjacencyLists includesKey: (document @ position)! !

!LDLinkField methodsFor: 'as yet unclassified' stamp: 'toma 8/14/2015 17:00'!
linkDocument: document at: position to: anotherDocument

	| links |

	links := adjacencyLists at: (document @ position) ifAbsentPut: [OrderedCollection new].
	links add: anotherDocument.! !

!LDLinkField methodsFor: 'as yet unclassified' stamp: 'toma 8/14/2015 17:01'!
linksFromDocument: document at: position

	^adjacencyLists at: (document @ position) ifAbsent: {}.

! !

Object subclass: #LDMarkovChain
	instanceVariableNames: 'transitionMatrix observationMatrix stateCount'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LDAS'!

!LDMarkovChain class methodsFor: 'as yet unclassified' stamp: 'toma 8/6/2015 15:51'!
newStates: numStates
	
	^(LDMarkovChain new)
		initializeStateCount: numStates;
		yourself! !

!LDMarkovChain class methodsFor: 'as yet unclassified' stamp: 'toma 8/6/2015 15:54'!
newUniformStates: numStates
	
	^(LDMarkovChain new)
		initializeUniformly: numStates;
		yourself! !

!LDMarkovChain methodsFor: 'as yet unclassified' stamp: 'toma 8/6/2015 15:44'!
initializeStateCount: numStates
	
	transitionMatrix := Matrix identity: numStates.
	observationMatrix := Matrix rows: numStates columns: numStates element: 0.
	stateCount := numStates ! !

!LDMarkovChain methodsFor: 'as yet unclassified' stamp: 'toma 8/6/2015 15:54'!
initializeUniformly: numStates
	
	transitionMatrix := Matrix rows: numStates columns: numStates
		element: (1.0 / numStates).
	observationMatrix := Matrix rows: numStates columns: numStates element: 0.
	stateCount := numStates ! !

!LDMarkovChain methodsFor: 'as yet unclassified' stamp: 'toma 8/6/2015 15:25'!
observeSequence: aCollection
	
	| state |
	state := 0.
	aCollection do: [ :each |
		(state ~= 0) ifTrue: [self observeTransitionFrom: state to: each].
		state := each.
	]! !

!LDMarkovChain methodsFor: 'as yet unclassified' stamp: 'toma 8/6/2015 15:22'!
observeTransitionFrom: aState to: anotherState
	
	observationMatrix at: aState at: anotherState put: (
		1 + (observationMatrix at: aState at: anotherState))! !

!LDMarkovChain methodsFor: 'as yet unclassified' stamp: 'toma 8/6/2015 15:49'!
stationary: iterations
	"Approximate the stationary probability vector using given number of iterations"

	| matrix |
	matrix := transitionMatrix.
	(1 to: iterations) do: [ :i | matrix := matrix +* matrix].
	^matrix atRow: 1! !

!LDMarkovChain methodsFor: 'as yet unclassified' stamp: 'toma 8/6/2015 15:32'!
transitionMatrix
	
	^transitionMatrix! !

!LDMarkovChain methodsFor: 'as yet unclassified' stamp: 'toma 8/6/2015 15:32'!
transitionMatrix: aMatrix
	
	transitionMatrix := aMatrix.
	stateCount := aMatrix rowCount.! !

!LDMarkovChain methodsFor: 'as yet unclassified' stamp: 'toma 8/6/2015 15:45'!
update
	"update transition probabilities from sampled observations"

	(1 to: stateCount) do: [ :state | | row sum |
		row := observationMatrix atRow: state.
		sum := row sum.
		transitionMatrix atRow: state put: (row collect: [:freq | (freq / sum) asFloat]).
	]
! !

Object subclass: #LDModel
	instanceVariableNames: 'perTopicFeatures perDocumentTopics topicPriors'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LDAS'!

!LDModel commentStamp: 'toma 8/6/2015 01:08' prior: 0!
This is the result of LDAllocator training. See LDAllocator documentation for examples.!

!LDModel methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 18:52'!
documentSimilarityOf: doc1 and: doc2
	
	^( (self topicsForDocumentAt: doc1)
		with: (self topicsForDocumentAt: doc2) collect: [ :i :k | i * k] ) sum! !

!LDModel methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 18:48'!
documentsForTopicAt: topic
	
	^perDocumentTopics atColumn: topic! !

!LDModel methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 18:53'!
featureSimilarityOf: word1 and: word2
	
	^( (self topicsForFeatureAt: word1)
		with: (self topicsForFeatureAt: word2) collect: [ :i :k | i * k] ) sum! !

!LDModel methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 18:45'!
featuresForTopicAt: topic

	^perTopicFeatures atRow: topic! !

!LDModel methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 18:57'!
initializeTopics: topics andDocuments: documents priors: priors
	
	perTopicFeatures := topics.
	perDocumentTopics := documents.
	topicPriors := priors.! !

!LDModel methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 18:45'!
topicsForDocumentAt: docid
	
	^perDocumentTopics atRow: docid! !

!LDModel methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 18:46'!
topicsForFeatureAt: word

	^perTopicFeatures atColumn: word! !

Object subclass: #LDMultinomial
	instanceVariableNames: 'probabilities'
	classVariableNames: 'Rand'
	poolDictionaries: ''
	category: 'LDAS'!

!LDMultinomial commentStamp: 'toma 8/6/2015 01:09' prior: 0!
This represents a multinomially distributed random variable and is used internally by LDAllocator to deal with multinomial distributions.!

!LDMultinomial class methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 17:58'!
initialize
	"LDMultinomial initialize"

	Rand := Random new.! !

!LDMultinomial class methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 18:37'!
normalized: observation
	^ (LDMultinomial new)
		observe: observation;
		yourself.! !

!LDMultinomial class methodsFor: 'as yet unclassified' stamp: 'toma 8/14/2015 14:40'!
rand
	^ Rand! !

!LDMultinomial class methodsFor: 'accessing' stamp: 'toma 8/14/2015 17:04'!
rand: anObject
	Rand := anObject! !

!LDMultinomial class methodsFor: 'as yet unclassified' stamp: 'toma 8/14/2015 17:05'!
raw: observation
	^ (LDMultinomial new)
		probabilities: observation;
		yourself.! !

!LDMultinomial methodsFor: 'as yet unclassified' stamp: 'toma 8/14/2015 17:10'!
+ aMultinomial

	"Even mixture of two multinomials"

	^LDMultinomial normalized: (self probabilities with: (aMultinomial probabilities) collect: [:a :b | a + b ])
	! !

!LDMultinomial methodsFor: 'as yet unclassified' stamp: 'toma 8/14/2015 17:09'!
++ aMultinomial

	"Non-normalizing addition. Faster in chained additions but requires subsequent normalization"

	^LDMultinomial raw: (self probabilities with: (aMultinomial probabilities) collect: [:a :b | a + b ])
	! !

!LDMultinomial methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 17:55'!
normalize
	| sum |
	
	sum := probabilities sum.
	probabilities := probabilities collect: [:p | p / sum]! !

!LDMultinomial methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 17:56'!
observe: frequencies
	
	probabilities := frequencies.
	self normalize! !

!LDMultinomial methodsFor: 'accessing' stamp: 'toma 8/14/2015 17:04'!
probabilities
	^ probabilities! !

!LDMultinomial methodsFor: 'accessing' stamp: 'toma 8/14/2015 17:04'!
probabilities: anObject
	probabilities := anObject! !

!LDMultinomial methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 18:04'!
sample
	"Draw a sample index from the distribution"
	| r |
	
	r := Rand next.
	probabilities withIndexDo: [ :p :i |
		r := r - p.
		(r < 0) ifTrue: [^ i]]! !

LDMultinomial subclass: #LDNullMultinomial
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LDAS'!

!LDNullMultinomial methodsFor: 'as yet unclassified' stamp: 'toma 8/14/2015 17:14'!
+ aMultinomial

	^aMultinomial! !

!LDNullMultinomial methodsFor: 'as yet unclassified' stamp: 'toma 8/14/2015 17:14'!
++ aMultinomial

	^aMultinomial! !

!LDNullMultinomial methodsFor: 'as yet unclassified' stamp: 'toma 8/14/2015 17:15'!
normalize
! !

!LDNullMultinomial methodsFor: 'as yet unclassified' stamp: 'toma 8/14/2015 17:16'!
probabilities

	self error: 'This is a dummy multinomial with weight 0 when added to another one'! !

!LDNullMultinomial methodsFor: 'as yet unclassified' stamp: 'toma 8/14/2015 17:16'!
sample

	self error: 'This is a dummy multinomial with weight 0 when added to another one'! !

Object subclass: #LDNeighborRandomField
	instanceVariableNames: 'documentSizes'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LDAS'!

!LDNeighborRandomField class methodsFor: 'as yet unclassified' stamp: 'toma 8/14/2015 14:56'!
forDocumentSizes: aCollection

	^(LDNeighborRandomField new)
		documentSizes: aCollection;
		yourself.! !

!LDNeighborRandomField methodsFor: 'accessing' stamp: 'toma 8/14/2015 14:50'!
documentSizes
	^ documentSizes! !

!LDNeighborRandomField methodsFor: 'accessing' stamp: 'toma 8/14/2015 14:50'!
documentSizes: anObject
	documentSizes := anObject! !

!LDNeighborRandomField methodsFor: 'as yet unclassified' stamp: 'toma 8/14/2015 14:55'!
markovBlanket: aPosition
	
	| document index |
	
	document := aPosition x.
	index := aPosition y.
	
	((documentSizes at: document) = 1) ifTrue: [ ^{} ].
	(index <= 1) ifTrue: [ ^ { document @ (index + 1) } ].
	(index >= (documentSizes at: document)) ifTrue: [ ^ { document @ (index - 1) } ].
	^ { document @ (index - 1). document @ (index + 1)  } 
	! !

Object subclass: #LDRandomField
	instanceVariableNames: 'adjacencyLists'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LDAS'!

!LDRandomField methodsFor: 'accessing' stamp: 'toma 8/14/2015 14:19'!
adjacencyLists
	^ adjacencyLists! !

!LDRandomField methodsFor: 'accessing' stamp: 'toma 8/14/2015 14:19'!
adjacencyLists: anObject
	adjacencyLists := anObject! !

!LDRandomField methodsFor: 'accessing' stamp: 'toma 8/14/2015 14:19'!
initialize
	adjacencyLists := Dictionary new.! !

!LDRandomField methodsFor: 'accessing' stamp: 'toma 8/14/2015 14:25'!
link: aPosition directionallyTo: anotherPosition
	
	(self markovBlanket: aPosition) add: anotherPosition! !

!LDRandomField methodsFor: 'accessing' stamp: 'toma 8/14/2015 14:25'!
link: aPosition to: anotherPosition

	self link: aPosition directionallyTo: anotherPosition.
	self link: anotherPosition directionallyTo: aPosition! !

!LDRandomField methodsFor: 'accessing' stamp: 'toma 8/14/2015 14:25'!
markovBlanket: aPosition

	^ self adjacencyLists at: aPosition ifAbsentPut: [ OrderedCollection new ]! !
LDMultinomial initialize!
